"""
Tests for the API (logged in user)
"""

import os
import json

import pytest

from datetime import datetime


@pytest.mark.parametrize(
    "limit, expected_number",
    [
        (None, 3),
        (2, 2),
    ],
)
def test_get_jobs(limit, expected_number, auth_client):
    """Tests get method for api/jobs with various query strings"""

    auth_client = auth_client[0]

    query_string = "api/jobs"

    if limit is not None:
        query_string += f"?limit={limit}"

    response = auth_client.get(query_string)
    jobs_received = response.json

    assert len(jobs_received) == expected_number, jobs_received


def test_get_job_by_id(auth_client):
    """API endpoint api/jobs/{jobID}"""
    auth_client = auth_client[0]

    response = auth_client.get("api/jobs/3")

    sub_time = datetime.fromisoformat("2016-08-29T09:12:00.000000+00:00").astimezone()
    sub_time = sub_time.strftime(format="%Y-%m-%d %H:%M")
    expected_response = {
        "flowchart_id": "ABCD",
        "id": 3,
        "path": "/Users/username/seamm/projects",
        "submitted": sub_time,
    }

    received = response.json

    for k in expected_response.keys():
        if k == "submitted":
            assert received[k] == expected_response[k]

    assert response.status_code == 200


def test_flowcharts_logged_in(auth_client):

    auth_client = auth_client[0]

    response = auth_client.get("api/flowcharts")

    assert len(response.json) == 1, "The response is" + str(response)
    assert response.status_code == 200


def test_get_flowchart_logged_in(auth_client):
    """
    Test for api/flowcharts/{flowchart_ID} when logged in and owner of flowchart
    """

    auth_client = auth_client[0]

    response = auth_client.get("api/flowcharts/100")

    assert response.status_code == 200


def test_get_cytoscape_logged_in(auth_client):
    """
    Test for api/flowcharts/{flowchart_ID}/cytoscape
    """
    auth_client = auth_client[0]

    response = auth_client.get("api/flowcharts/100/cytoscape")
    assert response.status_code == 201


def test_update_job(auth_client):
    """Check put method of api/jobs/{job_ID}"""

    csrf_token = auth_client[1]
    auth_client = auth_client[0]

    original_info = auth_client.get("api/jobs/1").json
    assert original_info["status"].lower() == "finished"
    response = auth_client.put(
        "api/jobs/1",
        data=json.dumps({"status": "submitted"}),
        headers={
            "Accept": "application/json",
            "Content-Type": "application/json",
            "X-CSRF-TOKEN": csrf_token,
        },
    )

    assert response.status_code == 201

    new_info = auth_client.get("api/jobs/1").json

    assert new_info["status"] == "submitted"


def test_update_project(auth_client):
    """Check put method of api/jobs/{job_ID}"""

    csrf_token = auth_client[1]
    auth_client = auth_client[0]
    original_info = auth_client.get("api/projects/1").json
    assert original_info["description"] is None

    response = auth_client.put(
        "api/projects/1",
        data=json.dumps({"description": "testing update"}),
        headers={
            "Accept": "application/json",
            "Content-Type": "application/json",
            "X-CSRF-TOKEN": csrf_token,
        },
    )

    assert response.status_code == 201

    new_info = auth_client.get("api/projects/1").json
    assert new_info["description"] == "testing update"


def test_add_project(auth_client):
    csrf_token = auth_client[1]
    auth_client = auth_client[0]

    project = {"name": "added_project"}

    response = auth_client.post(
        "api/projects",
        data=json.dumps(project),
        content_type="application/json",
        headers={"X-CSRF-TOKEN": csrf_token},
    )

    assert response.status_code == 201

    from seamm_dashboard import datastore

    assert os.path.exists(os.path.join(datastore, "projects", "added_project"))


def test_get_users(auth_client):
    """Check get method of api/users on authenticated client"""

    auth_client = auth_client[0]

    response = auth_client.get("api/users")

    assert response.status_code == 401


def test_get_projects(auth_client):
    """Check get method of api/projects on authenticated client"""

    auth_client = auth_client[0]

    response = auth_client.get("api/projects")

    assert response.status_code == 200

    assert len(response.json) == 2


def test_get_project(auth_client):
    """Check get method of api/projects on authenticated client"""

    auth_client = auth_client[0]

    response = auth_client.get("api/projects/100")

    assert response.status_code == 200


def test_get_project_jobs(auth_client):
    """Check get method of api/projects on authenticated client"""

    auth_client = auth_client[0]

    response = auth_client.get("api/projects/100/jobs")

    assert response.status_code == 200


def test_get_project_jobs_404(auth_client):
    """Check get method of api/projects on authenticated client"""

    auth_client = auth_client[0]

    response = auth_client.get("api/projects/1000/jobs")

    assert response.status_code == 404



"""
Tests for the front end
"""

from flask import url_for

import pytest
import requests
import os
import platform
import sys
import time

import urllib.parse

from selenium.webdriver.common.by import By
from selenium.webdriver.support.ui import WebDriverWait
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.common.by import By

# Needed for Python 3.8 since default for OS/X changed to spawn rather than
# fork.
if sys.version_info >= (3, 8) and platform.system() == "Darwin":
    import multiprocessing

    multiprocessing.set_start_method("fork")


@pytest.mark.usefixtures("live_server")
class TestLiveServer:
    """
    This uses selenium and the 'live_server' fixture which comes with
    pytest-flask.
    """

    @property
    def base_url(self):
        return url_for("main.index", _external=True)

    def log_in(self, chrome_driver, username="sample_user", password="sample_password"):
        """
        Function to log sample user in for testing.
        """
        login_url = f"{self.base_url}/login"
        chrome_driver.get(login_url)

        username_field = chrome_driver.find_element(By.ID, "username")
        username_field.send_keys(username)

        username_field = chrome_driver.find_element(By.ID, "password")
        username_field.send_keys(password)

        button = chrome_driver.find_element(By.ID, "submit")
        button.click()

        # For some reason we need to navigate away from the main page
        # before loading the page of interest or we time out
        chrome_driver.get(f"{self.base_url}api/status")

    def log_out(self, chrome_driver):
        """
        Function to make sure we are logged out
        """

        logout_url = f"{self.base_url}/logout"

        chrome_driver.get(logout_url)

    @pytest.mark.parametrize("logged_in", [True, False])
    def test_main_view(self, app, chrome_driver, logged_in):

        if logged_in:
            self.log_in(chrome_driver)
            # Should have three finished jobs, 1 flowchart, and 2 projects when
            # logged in.
            expected_values = "3 0 3 1 2".split()

        else:
            # Make sure we are logged out
            self.log_out(chrome_driver)
            # Should have one public job and nothing else.
            expected_values = "1 0 1 0 0".split()

        chrome_driver.get(self.base_url)
        ui_view = chrome_driver.find_element(By.ID, "ui-view")
        displayed_values = ui_view.find_elements(By.CLASS_NAME, "text-value")

        displayed_values = [x for x in displayed_values if x != ""]

        for i, value in enumerate(displayed_values):
            assert expected_values[i] == value.get_attribute("innerHTML")

    @pytest.mark.parametrize(
        "list_type, num_columns, num_rows, logged_in",
        [
            ("jobs", 7, 4, True),
            ("jobs", 7, 2, False),
            ("flowcharts", 5, 2, True),
            ("flowcharts", 5, 2, False),
            ("projects", 4, 3, True),
            ("projects", 4, 2, False),
        ],
    )
    def test_list_views(
        self, app, chrome_driver, list_type, num_columns, num_rows, logged_in
    ):

        # log in or log out
        if logged_in:
            self.log_in(chrome_driver)
        else:
            self.log_out(chrome_driver)

        get_url = f"{self.base_url}#{list_type}"

        # Get twice just to make sure this request goes through.
        chrome_driver.get(get_url)
        chrome_driver.get(get_url)

        if list_type == "projects":
            # Default view is card - switch to list.
            button = chrome_driver.find_element(By.ID, "toggle-list")
            button.click()

        # chrome_driver.get_screenshot_as_file(f"{list_type}_{logged_in}.png")

        # Get the jobs table. Will want to wait for this to be loaded,
        # of course.
        jobs_table = WebDriverWait(chrome_driver, 20).until(
            EC.presence_of_element_located((By.ID, list_type))
        )

        # Check table dimensions.
        table_headings = jobs_table.find_elements(By.TAG_NAME, "th")
        table_rows = jobs_table.find_elements(By.TAG_NAME, "tr")

        assert len(table_headings) == num_columns
        assert len(table_rows) == num_rows

        # Check the response code of the links in the table.
        table_links = jobs_table.find_elements(By.CLASS_NAME, "nav-link")
        for link in table_links:
            # This is just a thing we have to do because of the way 'nav-links'
            # are handled in the app.
            important_url = link.get_attribute("href")[len(self.base_url) :]
            actual_url = f"{self.base_url}/#{important_url}"
            response = requests.get(actual_url)
            assert response.status_code == 200

        # If we're not logged in we shouldn't see links (except for the public job)
        if not logged_in:
            if list_type == "jobs":
                assert len(table_links) == 1
            else:
                assert len(table_links) == 0

        # chrome_driver.get_screenshot_as_file(F'{list_type}_screenshot.png')

    @pytest.mark.xfail
    def test_job_report_file_tree(self, app, chrome_driver, project_directory):
        """
        Test to make sure file tree loads with correct number of elements.
        """
        # Have to log in for this test
        self.log_in(chrome_driver)

        # Get page with chromedriver.
        chrome_driver.get(f"{self.base_url}#jobs/1")

        # Set up samples for comparison - we need the location of the job
        # which is in a temporary directory
        test_dir = os.path.realpath(os.path.join(project_directory, "Job_000001"))

        num_files = len(os.listdir(test_dir))

        # Get the file tree. Wait for a specific element to load so we know the
        # tree is loaded.
        file_tree = chrome_driver.find_element(By.ID, "js-tree")

        test_file = os.path.realpath(
            os.path.join(project_directory, "Job_000001", "job.out")
        )

        test_file_id = urllib.parse.quote(test_file, safe="") + "_anchor"

        WebDriverWait(chrome_driver, 20).until(
            EC.presence_of_element_located((By.ID, test_file_id))
        )

        # Now get components.
        js_tree_contents = file_tree.find_elements(By.TAG_NAME, "li")

        num_files_in_tree = len(js_tree_contents)

        assert num_files_in_tree == num_files + 1

    def test_job_report_file_content(self, app, chrome_driver, project_directory):
        """
        Test to click file and make sure it is loaded into div.
        """
        # Have to log in for this
        self.log_in(chrome_driver)

        # Set up sample file for comparison.
        test_file = os.path.realpath(
            os.path.join(project_directory, "Job_000001", "job.out")
        )

        with open(test_file) as f:
            file_contents = f.read()
            file_contents_split = file_contents.split()

        test_file_id = urllib.parse.quote(test_file, safe="") + "_anchor"

        chrome_driver.get(f"{self.base_url}#jobs/1")

        # Initially, there should be nothing in the text box.

        initial_displayed_text = (
            WebDriverWait(chrome_driver, 20)
            .until(EC.presence_of_element_located((By.ID, "file-content")))
            .text
        )

        # Get a link for a file and click on it.
        job_link = WebDriverWait(chrome_driver, 20).until(
            EC.presence_of_element_located((By.ID, test_file_id))
        )
        job_link.click()

        # Give time to load
        time.sleep(1.25)

        # When clicked, file text should be displayed in the div.
        displayed_text = chrome_driver.find_element(By.ID, "file-content").text

        displayed_text_list = displayed_text.split()

        # Splitting on whitespace and rejoining let's us compare the file
        # contents without worrying about how whitespace is handled.
        assert initial_displayed_text == "", "initial displayed text error"
        assert " ".join(displayed_text_list) == " ".join(
            file_contents_split
        ), "displayed text error"

    def test_job_report_file_content_refresh(
        self, app, chrome_driver, project_directory
    ):
        """
        Test to click file and make sure it is loaded into div.
        """
        # Have to log in for this
        self.log_in(chrome_driver)

        # Set up sample file for comparison.
        test_file = os.path.realpath(
            os.path.join(project_directory, "Job_000001", "job.out")
        )

        with open(test_file) as f:
            file_contents = f.read()
            file_contents_split = file_contents.split()

        test_file_id = urllib.parse.quote(test_file, safe="") + "_anchor"

        chrome_driver.get(f"{self.base_url}#jobs/1")

        # Initially, there should be nothing in the text box.
        initial_displayed_text = (
            WebDriverWait(chrome_driver, 20)
            .until(EC.presence_of_element_located((By.ID, "file-content")))
            .text
        )

        # Get a link for a file and click on it.
        job_link = WebDriverWait(chrome_driver, 20).until(
            EC.presence_of_element_located((By.ID, test_file_id))
        )
        job_link.click()

        time.sleep(1)

        # When clicked, file text should be displayed in the div.
        displayed_text = chrome_driver.find_element(By.ID, "file-content").text

        displayed_text_list = displayed_text.split()

        # Splitting on whitespace and rejoining let's us compare the file
        # contents without worrying about how whitespace is handled.
        assert initial_displayed_text == ""
        assert " ".join(displayed_text_list) == " ".join(
            file_contents_split
        ), "initial load failed."

        # Update file on disk
        with open(test_file, "a+") as f:
            f.write("Appending this line")

        # Update expected text
        file_contents_split += "Appending this line".split()

        # Click refresh button
        refresh_button = chrome_driver.find_element(By.ID, "refresh")
        refresh_button.click()
        time.sleep(1)

        # Check the new displayed text
        new_displayed_text = chrome_driver.find_element(By.ID, "file-content").text
        new_displayed_list = new_displayed_text.split()

        assert " ".join(new_displayed_list) == " ".join(file_contents_split)

    def test_job_report_file_content_resize(
        self, app, chrome_driver, project_directory
    ):
        """
        Test to make sure file content element resizes when next element is
        clicked.
        """

        # Make sure we're logged in
        self.log_in(chrome_driver)

        first_file = os.path.realpath(
            os.path.join(project_directory, "Job_000001", "job.out")
        )
        second_file = os.path.realpath(
            os.path.join(project_directory, "Job_000001", "flowchart.flow")
        )

        first_file_id = urllib.parse.quote(first_file, safe="") + "_anchor"
        second_file_id = urllib.parse.quote(second_file, safe="") + "_anchor"

        chrome_driver.get(f"{self.base_url}#jobs/1")

        # Get a link for a file and click on it.
        job_link = WebDriverWait(chrome_driver, 20).until(
            EC.presence_of_element_located((By.ID, first_file_id))
        )
        job_link.click()

        flowchart_link = WebDriverWait(chrome_driver, 20).until(
            EC.presence_of_element_located((By.ID, second_file_id))
        )
        flowchart_link.click()

        # File content div should not be displayed if another div is clicked on.
        assert not chrome_driver.find_element(By.ID, "file-content").is_displayed()

    def test_admin_views_logged_out(self, app, chrome_driver):
        """
        Make sure admin views are protected
        """

        # Make sure we're logged out
        self.log_out(chrome_driver)

        # Try to access admin views - manage_users page
        chrome_driver.get(f"{self.base_url}/admin/manage_users")

        header = chrome_driver.find_elements(By.TAG_NAME, "h1")[0]

        assert header.text == "401"

        # Try to access admin views - create users page
        chrome_driver.get(f"{self.base_url}/admin/create_user")

        header = chrome_driver.find_elements(By.TAG_NAME, "h1")[0]

        assert header.text == "401"

    def test_admin_create_user(self, app, chrome_driver):

        # Make sure we're logged in as admin
        self.log_in(chrome_driver, username="admin_user", password="iamadmin")

        # Try to access admin views - manage_users page
        chrome_driver.get(f"{self.base_url}/admin/create_user")

        username_field = chrome_driver.find_element(By.ID, "username")
        username_field.send_keys("new_user")

        password_field = chrome_driver.find_element(By.ID, "password")
        password_field.send_keys("test_password")

        password2_field = chrome_driver.find_element(By.ID, "password2")
        password2_field.send_keys("test_password")

        firstname_field = chrome_driver.find_element(By.ID, "first_name")
        firstname_field.send_keys("FirstName")

        lastname_field = chrome_driver.find_element(By.ID, "last_name")
        lastname_field.send_keys("LastName")

        email_field = chrome_driver.find_element(By.ID, "email")
        email_field.send_keys("email@email.com")

        button = chrome_driver.find_element(By.ID, "submit")
        button.click()

        # Check that alert is found. If not found will result in error.
        chrome_driver.find_element_by_class_name("alert-success")

        table_rows = chrome_driver.find_element(By.ID, "users").find_elements(
            By.TAG_NAME, "tr"
        )

        # chrome_driver.get_screenshot_as_file(f'user_table.png')

        assert len(table_rows) == 7

    def test_admin_edit_user(self, app, chrome_driver):

        # Make sure we're logged in as admin
        self.log_in(chrome_driver, username="admin_user", password="iamadmin")

        # Try to access admin views - manage_users page
        chrome_driver.get(f"{self.base_url}/admin/manage_user/1")

        # chrome_driver.get_screenshot_as_file(f'updated_user.png')

        edit_button = chrome_driver.find_elements(
            By.CSS_SELECTOR, "#user-information-button .btn"
        )[0]
        edit_button.click()

        # chrome_driver.get_screenshot_as_file(f'updated_user_after_after_click.png')

        # with open('page_source.html', 'w+') as f:
        #    f.write(chrome_driver.page_source)

        firstname_field = chrome_driver.find_element(By.ID, "first_name")
        firstname_field.send_keys("FirstName")

        lastname_field = chrome_driver.find_element(By.ID, "last_name")
        lastname_field.send_keys("LastName")

        email_field = chrome_driver.find_element(By.ID, "email")
        email_field.send_keys("changed_address@email.com")

        # chrome_driver.get_screenshot_as_file(f'updated_user.png')

        button = chrome_driver.find_element(By.ID, "submit")
        button.click()

        chrome_driver.find_element(By.CLASS_NAME, "alert-success")
        # chrome_driver.get_screenshot_as_file(f'user_table.png')



#!/usr/bin/env python
from seamm_dashboard import create_app, options


def run():
    app = create_app()
    # app.run(debug=True, use_reloader=True)
    app.run(debug=False, use_reloader=False, port=options["port"])


if __name__ == "__main__":
    run()



"""
Tests for a visitor
"""


def test_get_jobs(visitor_client):
    """API endpoint api/jobs"""
    auth_client = visitor_client[0]
    response = auth_client.get("api/jobs")

    # Two because there is one public job, and one job they have been added to,
    assert len(response.json) == 3, response.json[0]["id"]
    assert response.status_code == 200



"""Flask app configuration
"""
import os


class BaseConfig:

    _basedir = os.path.abspath(os.path.dirname(__file__))
    STATIC_FOLDER = "static"
    ADMINS = frozenset(["janash@vt.edu", "psaxe@vt.edu"])
    SECRET_KEY = "SecretKeyForSessionSigning"
    EDIT_SOFTWARE_SALT = "ThisIsAnotherSalt"
    THREADS_PER_PAGE = 8
    SQLALCHEMY_TRACK_MODIFICATIONS = False

    # email
    MAIL_SERVER = os.environ.get("MAIL_SERVER", "smtp.googlemail.com")
    MAIL_PORT = int(os.environ.get("MAIL_PORT", "587"))
    MAIL_USE_TLS = os.environ.get("MAIL_USE_TLS", "true").lower() in ["true", "on", "1"]
    MAIL_USERNAME = os.environ.get("MAIL_USERNAME", "userhere")
    MAIL_PASSWORD = os.environ.get("MAIL_PASSWORD", "passhere")
    MAIL_SUBJECT_PREFIX = "[MolSSI Molecular Software DB]"
    MAIL_SENDER = "MolSSI Molecular DB Admin <info@molssi.org>"
    APP_ADMIN = os.environ.get("APP_ADMIN", "janash@vt.edu")
    EMAIL_CONFIRMATION_ENABLED = False

    # Client-side config
    API_RESULTS_PER_PAGE = 5

    # Jinia template
    REPLACE_NONE = ""
    GOOGLE_ANALYTICS_GTAG = "UA-116673029-1"
    GOOGLE_ANALYTICS_GTAG_submit = "UA-116673029-2"
    WTF_CSRF_ENABLED = True  # it's true by default, important to prevent CSRF attacks


class DevelopmentConfig(BaseConfig):
    _basedir = os.path.abspath(os.path.dirname(__file__))
    DEBUG = True
    TESTING = False
    WORDPRESS_DOMAIN = "http://localhost:8888"
    API_DOMAIN = "http://localhost:5000"
    SQLALCHEMY_DATABASE_URI = "sqlite:///" + os.path.join(
        _basedir, "data", "projects", "molssi_jobstore.db"
    )


class TestingConfig(BaseConfig):
    import tempfile

    DEBUG = True
    TESTING = True
    API_DOMAIN = "http://localhost:4000"
    SQLALCHEMY_DATABASE_URI = "sqlite:///:memory:"
    initialize = True


class ProductionConfig(BaseConfig):
    DEBUG = False
    TESTING = False
    WORDPRESS_DOMAIN = "http://molssi.org"
    API_DOMAIN = "http://api.molssi.org"
    MONGODB_SETTINGS = {
        "host": "mongodb://user:pass@xyz.mlab.com:27163/resources_website",
    }


config = {
    "development": DevelopmentConfig,
    "testing": TestingConfig,
}



"""
API calls for projects
"""

import logging
import os
import shutil

from pathlib import Path

from flask import Response
from flask_jwt_extended import jwt_required

from seamm_dashboard import datastore, db
from seamm_datastore.database.models import Project
from seamm_datastore.database.schema import JobSchema, ProjectSchema

from seamm_datastore.util import NotAuthorizedError


logger = logging.getLogger(__name__)

__all__ = [
    "get_projects",
    "add_project",
    "get_project",
    "get_project_jobs",
    "list_projects",
    "update_project",
    "delete_project",
]


@jwt_required(optional=True)
def update_project(id, body):
    from seamm_datastore.util import NotAuthorizedError

    try:
        project = Project.get_by_id(id, permission="update")
    except NotAuthorizedError:
        return Response(status=401)

    if project is None:
        return Response(status=404)

    project.update(id, **body)

    return Response(status=201)


@jwt_required(optional=True)
def delete_project():
    try:
        project = Project.get_by_id(id, permission="delete")
    except NotAuthorizedError:
        return Response(status=401)

    if not project:
        return Response(status=404)

    from seamm_dashboard import datastore

    path = project.path

    # Ensure that job path is absolute
    # and that it shares a base directory
    # with the datastore
    if not os.path.isabs(path) or os.path.commonprefix([datastore, path]) != datastore:
        return Response(status=401)

    path = project.path
    project_path = Path(path)

    # Remove job files if they exist
    if project_path.exists():
        shutil.rmtree(project_path)

    # Remove job info from DB
    db.session.delete(project)
    db.session.commit()

    return Response(status=200)


@jwt_required(optional=True)
def get_projects(
    permission="read",
    description=None,
    offset=None,
    limit=None,
    sort_by="id",
    order="asc",
):

    projects = Project.get(
        permission=permission,
        description=description,
        offset=offset,
        limit=limit,
        sort_by=sort_by,
        order=order,
    )

    projects = ProjectSchema(many=True).dump(projects)

    return projects, 200


@jwt_required(optional=True)
def add_project(body):
    """Add a new project

    Parameters
    ----------
    body : dict
        The description of the project.

    Returns
    -------
    str
        The project name

    The body contains:

        name : str
            The name of the project
        description : str
            A longer description of the project.
        owner : str
            The owner of the project. If not present defaults to the current logged in
             user.
    """
    logger.debug("Adding a project. Items in the body are:")
    for key, value in body.items():
        logger.debug("  {:15s}: '{}'".format(key, str(value)[:20]))
    name = body["name"]
    description = body["description"] if "description" in body else None

    # Create a directory for the project
    project_path = Path(datastore).expanduser() / "projects" / name

    project_path.mkdir(parents=True, exist_ok=True)

    Project.create(name=name, description=description, path=project_path)

    return {"name": name}, 201


@jwt_required(optional=True)
def get_project(id):
    """
    Function for api endpoint api/projects/{id}

    Parameters
    ----------
    id : the ID of the project to return
    """

    try:
        id = int(id)
    except ValueError:
        return Response(status=400)

    try:
        project = Project.get_by_id(id)
    except NotAuthorizedError:
        return Response("You are not authorized to view this content.", status=401)

    if project is None:
        return Response(status=404)

    project_schema = ProjectSchema(many=False)
    return project_schema.dump(project), 200


@jwt_required(optional=True)
def get_project_jobs(id):
    """
    Function for api endpoint api/projects/{id}/jobs. Get jobs associated with a
    project.

    Parameters
    ----------
    id : the ID of the project to return
    """

    try:
        id = int(id)
    except ValueError:
        return Response(status=400)

    try:
        project = Project.get_by_id(id, permission="read")
    except NotAuthorizedError:
        return Response("You are not authorized to view this content.", status=401)

    if project is None:
        return Response(status=404)

    jobs_schema = JobSchema(many=True)

    return jobs_schema.dump(project.jobs), 200


@jwt_required(optional=True)
def list_projects(
    permission="read",
    description=None,
    offset=None,
    limit=None,
    sort_by="id",
    order="asc",
):

    projects = Project.get(
        permission=permission,
        description=description,
        offset=offset,
        limit=limit,
        sort_by=sort_by,
        order=order,
        only="name",
    )
    project_list = [x.name for x in projects]
    return project_list, 200



from flask import Blueprint

main = Blueprint("main", __name__)

from . import views, errors  # noqa: F401, E402

# from ..models.users import Permission


# @main.app_context_processor
# def inject_permissions():
#     return dict(Permission=Permission)



"""
Routes for REST authentication
"""

from datetime import timedelta

from flask import jsonify, Response, make_response, request, redirect, url_for
from flask_jwt_extended import (
    create_access_token,
    create_refresh_token,
    set_access_cookies,
    set_refresh_cookies,
    jwt_required,
    get_jwt_identity,
    unset_jwt_cookies,
)

from seamm_datastore.database.models import User
from seamm_datastore.database.schema import UserSchema

from seamm_dashboard import jwt

__all__ = ["get_auth_token", "refresh_auth_token", "remove_auth_token"]


def create_tokens(user):
    """
    Create a token for a given user object. Not an api endpoint
    """

    # Access tokens expire after an hour.
    # Refresh tokens expire after a day.
    exp_time = timedelta(hours=1)
    exp2_time = timedelta(days=1)

    user_schema = UserSchema(many=False)
    user = user_schema.dump(user)
    access_token = create_access_token(identity=user, expires_delta=exp_time)
    refresh_token = create_refresh_token(identity=user, expires_delta=exp2_time)

    return access_token, refresh_token


def get_auth_token(body):
    """
    Endpoint for api/auth/token
    """

    username = body.pop("username")
    password = body.pop("password")

    user = User.query.filter_by(username=username).one_or_none()

    if not user:
        return jsonify({"msg": f"User {username} not found."}), 400

    if user.verify_password(password):

        access_token, refresh_token = create_tokens(user)

        resp = Response({"login": True})
        set_access_cookies(resp, access_token)
        set_refresh_cookies(resp, refresh_token)
        return resp, 200
    else:
        return jsonify({"msg": "Incorrect password."}), 400


def remove_auth_token():
    resp = make_response({"logout": True})
    unset_jwt_cookies(resp)
    return resp, 200


@jwt_required(refresh=True)
def refresh_auth_token():
    """
    Route for refreshing the access token.
    """

    exp_time = timedelta(hours=1)

    # Create the new access token
    current_user = get_jwt_identity()
    access_token = create_access_token(
        identity=current_user, expires_delta=exp_time, fresh=False
    )

    # Set the JWT access cookie in the response
    resp = Response({"refresh": True})
    set_access_cookies(resp, access_token)
    return resp, 200


@jwt.unauthorized_loader
def needed_token_callback(_):
    if "api" in request.url:
        return (
            jsonify(
                {
                    "status": 401,
                    "sub_status": 42,
                    "msg": "You are missing a token, please login.",
                }
            ),
            401,
        )

    else:
        return redirect(url_for("auth.login"))


@jwt.expired_token_loader
def my_expired_token_callback(jwt_header, expired_token):

    token_type = expired_token["type"]
    return (
        jsonify(
            {
                "status": 401,
                "sub_status": 42,
                "msg": "The {} token has expired".format(token_type),
            }
        ),
        401,
    )



"""
Admin views
"""

from copy import deepcopy

from flask import render_template, redirect, url_for, flash, Response, request

from flask_jwt_extended import (
    jwt_required,
    get_current_user,
)

from .forms import (
    CreateUserForm,
    ManageUserFormAdmin,
    EditGroupForm,
    DeleteUserForm,
    DeleteGroupForm,
    _validate_username,
    _validate_email,
    _validate_group,
    _validate_user_delete,
)

from wtforms import BooleanField

from . import admin

from seamm_dashboard import db, authorize
from seamm_datastore.database.models import (
    Role,
    Group,
    User,
    Project,
    GroupProjectAssociation,
    UserProjectAssociation,
)
from seamm_dashboard.routes.api.users import _process_user_body


def _process_user_permissions(filled_form, user):

    # Empty dictionary containing all entries
    permissions_dict = {
        int(k.split("_")[1]): []
        for k in filled_form.data.keys()
        if "specialproject" in k
    }

    # Get just the checked fields
    specialproject_keys = [
        x
        for x in filled_form.data.keys()
        if "specialproject" in x
        if filled_form.data[x] is True
    ]

    # Loop over checked fields and update dictionary containing all entries.
    # First collect permissions which were set in form
    # Build permissions dictionary (keys are project IDs,
    # values are permissions ("read", "update", etc))
    for key in specialproject_keys:
        split = key.split("_")
        project_id = int(split[1])
        permissions_dict[project_id].append(split[2])

    # Loop through permissions dictionary built in previous loop.
    # This is done after processing permissions because we need a full
    # list of the permissions for updating.
    for project_id, permission in permissions_dict.items():
        project = Project.query.filter_by(id=project_id).one()

        # Look to see if setting exists yet
        assoc = UserProjectAssociation.query.filter_by(
            entity_id=user.id, resource_id=project.id
        ).one_or_none()

        if not assoc:
            assoc = UserProjectAssociation(
                entity_id=user.id, resource_id=project.id, permissions=permission
            )
        else:
            assoc.permissions = permission

        db.session.add(assoc)
        db.session.commit()


def _bind_user_projects_to_form(form, projects, user=None, new=True):

    if not new and not user:
        raise ValueError(
            "A user must be given if the form is being created for an existing user."
        )

    project_names = []
    actions = ["read", "update", "create", "delete", "manage"]

    # Bind boolean fields for permission types
    for project in projects:
        field_name = f"specialproject_{project.id}"
        permissions = []
        if not new:
            assoc = UserProjectAssociation.query.filter_by(
                resource_id=project.id, entity_id=user.id
            ).one_or_none()

            if assoc:
                permissions = assoc.permissions

        for action in actions:
            checked = action in permissions
            setattr(form, f"{field_name}_{action}", BooleanField(default=checked))

        project_names.append({"name": project.name, "id": project.id})

    return form, project_names


def _bind_special_projects_to_form(form, projects, group=None, new=True):

    if not new and not group:
        raise ValueError(
            "A group must be given if the form is being created for an existing group."
        )

    project_names = []
    actions = ["read", "update", "create", "delete", "manage"]

    # Bind boolean fields for permission types
    for project in projects:
        field_name = f"specialproject_{project.id}"
        permissions = []
        if not new:
            assoc = GroupProjectAssociation.query.filter_by(
                resource_id=project.id, entity_id=group.id
            ).one_or_none()

            if assoc:
                permissions = assoc.permissions

        for action in actions:
            checked = action in permissions
            setattr(form, f"{field_name}_{action}", BooleanField(default=checked))

        project_names.append({"name": project.name, "id": project.id})

    return form, project_names


def _bind_owned_projects_to_form(form, projects, group=None, new=True):

    if not new and not group:
        raise ValueError(
            "A group must be given if the form is being created for an existing group."
        )

    project_names = []
    actions = ["read", "update", "create", "delete", "manage"]

    # Bind boolean fields for permission types
    for project in projects:
        field_name = f"ownedproject_{project.id}"
        permissions = project.group_permissions

        for action in actions:
            checked = action in permissions
            setattr(form, f"{field_name}_{action}", BooleanField(default=checked))

        project_names.append({"name": project.name, "id": project.id})

    return form, project_names


def _process_group_form_data(form):

    # Process form data
    # Get users specified on form from db
    users = User.query.filter(User.username.in_(form.data["group_members"])).all()

    group = Group.query.filter_by(name=form.data["group_name"]).one_or_none()

    if group is None:
        # Create new group containing users
        group = Group(name=form.data["group_name"], users=users)
    else:
        group.users = users

    db.session.add(group)
    db.session.commit()

    permissions_dict = {}

    specialproject_keys = [
        x for x in form.data.keys() if "specialproject" in x if form.data[x] is True
    ]

    # First collect permissions which were set in form
    for key in specialproject_keys:
        split = key.split("_")
        project_id = int(split[1])

        try:
            permissions_dict[project_id].append(split[2])
        except KeyError:
            permissions_dict[project_id] = [split[2]]

    for project_id, permission in permissions_dict.items():
        project = Project.query.filter_by(id=project_id).one()

        # Look to see if setting exists yet
        assoc = GroupProjectAssociation.query.filter_by(
            entity_id=group.id, resource_id=project.id
        ).one_or_none()

        if not assoc:
            assoc = GroupProjectAssociation(
                entity_id=group.id, resource_id=project.id, permissions=permission
            )
        else:
            assoc.permissions = permission

        db.session.add(assoc)
        db.session.commit()

    # Set owned project permissions
    owned_projects = Project.query.filter_by(group_id=group.id)
    for project in owned_projects:
        # Zero permissions
        perm = project.permissions
        perm["group"] = []
        project.set_permissions(perm)
        db.session.add(project)
        db.session.commit()

    ownedproject_keys = [
        x for x in form.data.keys() if "ownedproject" in x if form.data[x] is True
    ]

    # Update permissions
    for key in ownedproject_keys:
        split = key.split("_")
        project_id = int(split[1])
        permission = split[2]

        project = Project.query.filter_by(id=project_id).one()
        perm = deepcopy(project.permissions)
        perm["group"].append(permission)
        project.set_permissions(perm)
        db.session.add(project)
        db.session.commit()


@admin.route("/admin/manage_users")
@jwt_required(optional=True)
def manage_users():
    if not authorize.has_role("admin"):
        return render_template("401.html")
    return render_template("admin/manage_users.html")


@admin.route("/admin/manage_groups")
@jwt_required(optional=True)
def manage_groups():
    if not authorize.has_role("admin", "group manager"):
        return render_template("401.html")
    return render_template("admin/manage_groups.html")


@admin.route("/admin/create_group", methods=["GET", "POST"])
@jwt_required(optional=True)
def create_group():
    if not authorize.has_role("admin", "group manager"):
        return render_template("401.html")

    # We have to add these fields dynamically based on the projects in the database
    projects = Project.query.all()

    form_copy = deepcopy(EditGroupForm)
    form_copy, project_names = _bind_special_projects_to_form(form_copy, projects)

    form = form_copy()
    users = User.query.all()
    form.group_members.choices = [(user.username, user.username) for user in users]

    if form.validate_on_submit():

        _process_group_form_data(form)

        flash(f"The group {form.data['group_name']} has been successfully created")
        return render_template("admin/manage_groups.html")

    return render_template(
        "admin/create_group.html", form=form, special_project_names=project_names
    )


@admin.route("/admin/manage_group/<group_id>", methods=["GET", "POST"])
@jwt_required(optional=True)
def manage_group(group_id):
    # Permissions check
    if not authorize.has_role("admin", "group manager"):
        return render_template("401.html")

    # Get the group information
    group = Group.query.get(group_id)

    if not group:
        return render_template("404.html")

    # We have to add project fields dynamically based on the projects in the database

    # Get projects owned by the group
    owned_projects = Project.query.filter_by(group_id=group.id).all()

    # Get all projects in db
    projects = Project.query.all()

    # special projects
    special_projects = list(set(projects) - set(owned_projects))

    form_copy = deepcopy(EditGroupForm)
    form_copy, special_projects = _bind_special_projects_to_form(
        form_copy, special_projects, new=False, group=group
    )
    form_copy, owned_projects = _bind_owned_projects_to_form(
        form_copy, owned_projects, new=False, group=group
    )

    form = form_copy()
    users = User.query.all()
    form.group_members.choices = [(user.username, user.username) for user in users]

    # Set defaults
    if request.method == "GET":
        form.group_name.data = group.name
        form.group_members.data = [g.username for g in group.users]

    if request.method == "POST":
        try:
            form.group_name.validators.remove(_validate_group)
        except ValueError:
            pass

        # field has been attempted to be updated and we
        # must check the input
        if form.group_name.data != group.name:
            if Group.query.filter(Group.name == form.group_name.data).first():
                form.group_name.validators.append(_validate_group)

        # use validate instead of validate_on_submit so we can add our own validators
        # (lines above)
        if form.validate():

            _process_group_form_data(form)

            db.session.commit()
            flash(f"The group {form.data['group_name']} has been successfully updated.")
            return render_template("admin/manage_groups.html")

    return render_template(
        "admin/create_group.html",
        form=form,
        owned_project_names=owned_projects,
        special_project_names=special_projects,
        group_name=group.name,
        group_id=group.id,
    )


@admin.route("/admin/create_user", methods=["GET", "POST"])
@jwt_required(optional=True)
def create_user():

    if not authorize.has_role("admin"):
        return render_template("401.html")

    projects = Project.query.all()

    form = deepcopy(CreateUserForm)
    form, project_names = _bind_user_projects_to_form(form, projects=projects)
    form = form()
    form.groups.choices = [(g.name, g.name) for g in Group.query.all()]
    form.roles.choices = [(r.name, r.name) for r in Role.query.all()]

    if form.validate_on_submit():
        processed_form = _process_user_body(form.data)

        if isinstance(processed_form, Response):
            flash(
                "Creating the user failed because of problems with the input data. "
                "Please check the inputs and try again."
            )
            return redirect(url_for("admin.create_user"))

        else:
            db.session.add(processed_form)
            db.session.commit()

            _process_user_permissions(form, processed_form)

            flash(f"The user {form.data['username']} has been successfully created")
            return render_template("admin/manage_users.html")

    return render_template("admin/create_user.html", form=form, projects=project_names)


@admin.route("/admin/manage_user/<user_id>", methods=["GET", "POST"])
@jwt_required(optional=True)
def manage_user(user_id):

    # Permissions check
    if not authorize.has_role("admin"):
        return render_template("401.html")

    # Get the user information
    user = User.query.get(user_id)

    if not user:
        return render_template("404.html")

    projects = Project.query.all()

    form = deepcopy(ManageUserFormAdmin)
    form, project_names = _bind_user_projects_to_form(
        form, projects=projects, user=user, new=False
    )
    form = form()

    form.groups.choices = [(g.name, g.name) for g in Group.query.all()]
    form.roles.choices = [(r.name, r.name) for r in Role.query.all()]

    # Set defaults based on user
    if request.method == "GET":
        form.username.data = user.username
        form.groups.data = [g.name for g in user.groups]
        form.roles.data = [r.name for r in user.roles]
        form.first_name.data = user.first_name
        form.last_name.data = user.last_name
        form.email.data = user.email

    if request.method == "POST":

        # Each of these has to be tried - two separate try statements
        try:
            form.username.validators.remove(_validate_username)
        except ValueError:
            pass

        try:
            form.email.validators.remove(_validate_email)
        except ValueError:
            pass

        # field has been attempted to be updated and we
        # must check the input
        if form.username.data != user.username:
            if User.query.filter(User.username == form.username.data).first():
                form.username.validators.append(_validate_username)

        if form.email.data != user.email:
            if User.query.filter(User.email == form.email.data).first():
                form.email.validators.append(_validate_email)

        # use validate instead of validate_on_submit so we can add our own validators
        # (lines above)
        if form.validate():

            user = _process_user_body(form.data, original_user_data=user)
            db.session.add(user)
            db.session.commit()
            _process_user_permissions(form, user)
            flash(f"The user {form.data['username']} has been successfully updated.")
            return render_template("admin/manage_users.html")

    return render_template(
        "admin/create_user.html",
        form=form,
        username=user.username,
        user_id=user.id,
        projects=project_names,
    )


@admin.route("/admin/manage_user/<user_id>/delete", methods=["GET", "POST"])
@jwt_required(fresh=True)
def delete_user(user_id):
    # Permissions check
    if not authorize.has_role("admin"):
        return render_template("401.html")

    user_remove = User.query.filter(User.id == user_id).one()

    if user_remove.id == get_current_user().id:
        flash("You cannot remove your own account from the dashboard.")
        return render_template("admin/manage_users.html")

    form = DeleteUserForm()

    try:
        form.username.validators.remove(_validate_username)
    except ValueError:
        pass

    if request.method == "POST":

        specified_user = User.query.filter_by(username=form.username.data).one_or_none()

        if specified_user is None or str(specified_user.id) != user_id:
            form.username.validators.append(_validate_user_delete)
            flash("The input username did not match the requested user.")

        if form.validate():
            db.session.delete(specified_user)
            db.session.commit()
            flash(f"User {specified_user.username} removed from the dashboard.")

            return render_template("admin/manage_users.html")

    return render_template("admin/delete_user.html", form=form)


@admin.route("/admin/manage_group/<group_id>/delete", methods=["GET", "POST"])
@jwt_required(fresh=True)
def delete_group(group_id):
    # Permissions check
    if not authorize.has_role("admin", "group manager"):
        return render_template("401.html")

    form = DeleteGroupForm()

    if request.method == "POST":

        specified_group = Group.query.filter_by(name=form.group_name.data).one_or_none()

        if specified_group is None or str(specified_group.id) != group_id:
            form.group_name.validators.append(_validate_user_delete)
            flash("The input group name did not match the requested group.")

        if form.validate():
            db.session.delete(specified_group)
            db.session.commit()
            flash(f"Group {specified_group.name} removed from the dashboard.")

            return render_template("admin/manage_groups.html")

    return render_template("admin/delete_group.html", form=form)



"""
Tests for the API (public user)
"""
import json

import pytest


def test_get_protected_job(client):

    response = client.get("api/jobs/2")

    assert response.status_code == 401


def test_get_job_missing(client):
    """
    API endpoint api/jobs/{job_number}

    Test 404 response for job which does not exist

    """
    response = client.get("api/jobs/100")

    assert response.status_code == 404


def test_flowcharts_logged_out(client):
    """API endpoint for api/flowcharts"""

    response = client.get("api/flowcharts")

    assert len(response.json) == 0
    assert response.status_code == 200


def test_get_flowchart_logged_out(client):
    """
    API endpoint for api/flowcharts/{flowchart_ID}

    Get flowchart by ID
    """

    response = client.get("api/flowcharts/100")
    assert response.status_code == 401


def test_get_cytoscape(client):
    """
    API endpoint for api/flowcharts/{flowchart_ID}/cytoscape when not logged in.

    Get cytoscape representation of flowchart graph.
    """

    response = client.get("api/flowcharts/100/cytoscape")
    assert response.status_code == 401


def test_update_job_not_authenticated(client):
    """Check put method of api/jobs/{job_ID}"""

    response = client.put(
        "api/jobs/3",
        data=json.dumps({"status": "submitted"}),
        headers={"Accept": "application/json", "Content-Type": "application/json"},
    )

    assert response.status_code == 401


@pytest.mark.xfail
def test_add_job(client):
    """Check post method of api/jobs/"""
    # Ask Paul
    assert False


def test_delete_job(client):
    """Check delete method of api/jobs/{jobID}"""

    response = client.delete("api/jobs/2")

    assert response.status_code == 401


def test_get_users(client):
    """Check get method of api/users on unauthenticated client"""

    response = client.get("api/users")

    assert response.status_code == 401


def test_get_projects_unauthenticated(client):
    """Check get method of api/projects on unauthenticated client"""

    response = client.get("api/projects")

    assert response.status_code == 200

    assert len(response.json) == 0


def test_get_project_unauthenticated(client):
    """Check get method of api/projects on unauthenticated client"""

    response = client.get("api/projects/1")

    assert response.status_code == 401


def test_get_project_not_found(client):
    """Check get method of api/projects on unauthenticated client"""

    response = client.get("api/projects/3")

    assert response.status_code == 404



"""
API calls for jobs.
"""

import os
import shutil
from datetime import datetime, timezone
import fasteners
import json
import logging
from pathlib import Path
import re
import urllib.parse
import tempfile

from flask import send_from_directory, Response, request
from flask_jwt_extended import jwt_required

from seamm_dashboard import db, datastore, authorize, options
from seamm_datastore.database.models import Job
from seamm_datastore.database.schema import JobSchema

from seamm_datastore.util import NotAuthorizedError

logger = logging.getLogger("__file__")

__all__ = [
    "add_file_to_job",
    "get_jobs",
    "get_job",
    "get_job_files",
    "download_job_files",
    "add_job",
    "update_job",
    "delete_job",
]

file_icons = {
    "graph": "fas fa-chart-line",
    "csv": "fas fa-table",
    "flow": "fas fa-project-diagram",
    "other": "far fa-file-alt",
    "folder": "far fa-folder-open",
    "pdb": "fas fa-cubes",
    "mmcif": "fas fa-cubes",
    "cif": "fas fa-cubes",
}


@jwt_required(optional=True)
def get_jobs(
    permission="read",
    description=None,
    title=None,
    offset=None,
    limit=None,
    sort_by="id",
    order="asc",
):
    """
    Function for API endpoint /api/jobs
    """

    jobs = Job.get(
        permission=permission,
        description=description,
        title=title,
        offset=offset,
        limit=limit,
        sort_by=sort_by,
        order=order,
    )

    jobs = JobSchema(many=True).dump(jobs)

    return jobs


def get_job_id(filename):
    """Get the next job id from the given file.

    This uses the fasteners module to provide locking so that
    only one job at a time can access the file, so that the job
    ids are unique and monotonically increasing.
    """

    filename = os.path.expanduser(filename)

    lock_file = filename + ".lock"
    lock = fasteners.InterProcessLock(lock_file)
    locked = lock.acquire(blocking=True, timeout=5)

    if locked:
        if not os.path.isfile(filename):
            job_id = 1
            with open(filename, "w") as fd:
                fd.write("!MolSSI job_id 1.0\n")
                fd.write("1\n")
            lock.release()
        else:
            with open(filename, "r+") as fd:
                line = fd.readline()
                pos = fd.tell()
                if line == "":
                    lock.release()
                    raise EOFError("job_id file '{}' is empty".format(filename))
                line = line.strip()
                match = re.fullmatch(r"!MolSSI job_id ([0-9]+(?:\.[0-9]+)*)", line)
                if match is None:
                    lock.release()
                    raise RuntimeError(
                        "The job_id file has an incorrect header: {}".format(line)
                    )
                line = fd.readline()
                if line == "":
                    lock.release()
                    raise EOFError("job_id file '{}' is truncated".format(filename))
                try:
                    job_id = int(line)
                except TypeError:
                    raise TypeError(
                        "The job_id in file '{}' is not an integer: {}".format(
                            filename, line
                        )
                    )
                job_id += 1
                fd.seek(pos)
                fd.write("{:d}\n".format(job_id))
    else:
        raise RuntimeError("Could not lock the job_id file '{}'".format(filename))

    return job_id


@jwt_required(optional=True)
def add_job(body):
    """Add a new job to the queue.

    Parameters
    ----------
    body : dict
        The description of the job.

    Returns
    -------
    The job id (integer)

    The body contains:

        flowchart : str
            The flowchart
        project : [str]
            The projects associated with this job.
        title : str
            The title of the job (<100 chars)
        description : str
            A longer description of the job.
    """

    logger.debug("Adding a job. Items in the body are:")
    for key, value in body.items():
        logger.debug("  {:15s}: {}".format(key, str(value)[:20]))

    flowchart = body["flowchart"]
    project_names = [body["project"]]
    title = body["title"]
    description = body["description"]
    if "parameters" in body:
        parameters = body["parameters"]
    else:
        parameters = {}

    # Get the unique ID for the job...
    if options["job_id_file"] is None:
        job_id_file = os.path.join(datastore, "job.id")
    else:
        job_id_file = options["job_id_file"]
    job_id = get_job_id(job_id_file)

    # Create the directory and write the flowchart to it.
    project_path = Path(datastore).expanduser() / "projects" / project_names[0]
    directory = project_path / "Job_{:06d}".format(job_id)
    logger.info("Writing job files to " + str(directory))
    directory.mkdir(parents=True, exist_ok=True)

    path = directory / "flowchart.flow"
    with path.open("w") as fd:
        fd.write(flowchart)
    flowchart_file = str(path)

    # Write the json data file for the job
    data = {
        "data_version": "1.0",
        "command line": "",
        "title": body["title"],
        "working directory": str(directory),
        "state": "submitted",
        "projects": project_names,
        "datastore": datastore,
        "job id": job_id,
        "submitted": datetime.now(timezone.utc).isoformat(),
    }
    path = directory / "job_data.json"
    with path.open("w") as fd:
        fd.write("!MolSSI job_data 1.0\n")
        json.dump(data, fd, sort_keys=True, indent=3)

    job = Job.create(
        job_id,
        path=str(directory),
        flowchart_filename=flowchart_file,
        project_names=project_names,
        title=title,
        description=description,
        parameters=parameters,
    )

    db.session.add(job)
    db.session.commit()

    job = JobSchema.dump(job)

    return job


@jwt_required(optional=True)
def get_job(id):
    """
    Function for api endpoint api/jobs/{id}

    Parameters
    ----------
    id : the ID of the job to return
    """

    try:
        id = int(id)
    except ValueError:
        return Response(status=400)

    try:
        job = Job.get_by_id(id)
    except NotAuthorizedError:
        return Response("You are not authorized to view this content.", status=401)

    if job is None:
        return Response(status=404)

    job_schema = JobSchema(many=False)
    return job_schema.dump(job), 200


@jwt_required(optional=True)
def update_job(id, body):
    """
    Function to update jobs - endpoint api/jobs/{id}

    Parameters
    ----------
    id : int
        The ID of the job to update

    body : json
        The job information to update.
    """

    from seamm_datastore.util import NotAuthorizedError

    try:
        job = Job.get_by_id(id, permission="update")
    except NotAuthorizedError:
        return Response(status=401)

    if job is None:
        return Response(status=404)

    job.update(id=id, **body)

    return Response(status=201)


@jwt_required(optional=True)
@authorize.has_role("admin")
def delete_job(id):
    """
    Function for delete method of api endpoint api/jobs/{id}

    This api route removes the job from the DB and deletes the associated job files
    from disk

    Parameters
    ----------
    id : int
        The ID of the job to delete

    Returns
    -------
    status : int
        Response code for operation. 200 = successful, 404 = job not found.
    """

    try:
        job = Job.get_by_id(id, permission="delete")
    except NotAuthorizedError:
        return Response(status=401)

    if not job:
        return Response(status=404)

    from seamm_dashboard import datastore

    path = job.path

    # Ensure that job path is absolute
    # and that it shares a base directory
    # with the datastore
    if not os.path.isabs(path) or os.path.commonprefix([datastore, path]) != datastore:
        return Response(status=401)

    job_path = Path(path)

    # Remove job files if they exist
    if job_path.exists():
        shutil.rmtree(job_path)

    # Remove job info from DB
    db.session.delete(job)
    db.session.commit()

    return Response(status=200)


@jwt_required(optional=True)
def get_job_files(id):
    """
    Function for get method of api endpoint api/jobs/{id}/files.

    Parameters
    ----------
    id : int
        the ID of the job to return
    """
    try:
        job = Job.get_by_id(id)
    except NotAuthorizedError:
        return Response(status=401)

    js_tree = []

    path = job.path

    base_dir = os.path.split(path)[1]

    js_tree.append(
        {
            "id": path,
            "parent": "#",
            "text": base_dir,
            "state": {
                "opened": "true",
                "selected": "true",
            },
            "icon": file_icons["folder"],
        }
    )

    for root, dirs, files in os.walk(path):
        parent = root
        safe = parent.replace(job.path, "")[1:]

        for name in sorted(files):

            extension = name.split(".")[-1]

            encoded_path = urllib.parse.quote(os.path.join(root, name), safe="")
            safe_encode = urllib.parse.quote(os.path.join(safe, name), safe="")

            js_tree.append(
                {
                    "id": encoded_path,
                    "parent": parent,
                    "text": name,
                    "a_attr": {
                        "href": f"api/jobs/{id}/files/download?filename={safe_encode}",
                        "class": "file",
                    },
                    "icon": [
                        file_icons[extension]
                        if extension in file_icons.keys()
                        else file_icons["other"]
                    ][0],
                }
            )

        for name in sorted(dirs):
            js_tree.append(
                {
                    "id": os.path.join(root, name),
                    "parent": parent,
                    "text": name,
                    "icon": file_icons["folder"],
                }
            )
    return js_tree


@jwt_required(optional=True)
def download_job_files(id, filename=None):
    """
    Function for get method of api endpoint api/jobs/{id}/files/download.

    Without a query, will return zip file of job directory.
    With query, returns specified file.

    Parameters
    ----------
    id : int
        the ID of the job to return
    """

    try:
        job = Job.get_by_id(id)
    except NotAuthorizedError:
        return Response(status=401)

    path = job.path
    path, job_directory = os.path.split(path)

    if filename is None:
        # Create zip file in temporary directory and send.
        tmpdir = tempfile.mkdtemp()
        tmpzip = os.path.join(tmpdir, job_directory)
        shutil.make_archive(f"{tmpzip}", "zip", job.path)
        return send_from_directory(
            tmpdir, path=f"{job_directory}.zip", as_attachment=True
        )
    else:
        if "../" in filename:
            return Response(status=401)
        unencoded_path = urllib.parse.unquote(filename)
        return send_from_directory(job.path, path=unencoded_path, as_attachment=True)


@jwt_required(optional=True)
def add_file_to_job(body, id=None):
    """Add a new file to a job

    Parameters
    ----------
    body : dict
        The description of the file.

    Returns
    -------
    The job id (integer)
    """
    job_info, status = get_job(id)
    root = Path(job_info["path"])

    file_data = request.files["file"]
    filename = file_data.filename

    if filename[0:4] == "job:":
        filename = filename[4:]

    path = root / filename
    path.parent.mkdir(parents=True, exist_ok=True)

    logger.info(f"Saving file to {root / filename}")
    file_data.save(root / filename)

    return {"path": str(root / filename)}, 201



from flask_wtf import FlaskForm
from flask_wtf.file import FileField
from wtforms import StringField, TextAreaField, SubmitField

from wtforms.validators import DataRequired


class EditJob(FlaskForm):
    name = StringField("Job Title")
    notes = TextAreaField("Description")
    submit = SubmitField("Update Job")


class ImportJob(FlaskForm):
    outfile = FileField("job_data.json File", validators=[DataRequired()])
    title = StringField("Job Title")
    submit = SubmitField("Import Job")



import argparse
from pathlib import Path
import sys

import seamm_util

parser = seamm_util.seamm_parser()

# Options for the dashboard
parser.add_argument_group(
    "SEAMM",
    "dashboard options",
    "The options for the dashboard",
)

parser.add_argument(
    "SEAMM",
    "--port",
    group="dashboard options",
    default=5000,
    type=int,
    help="the port to use",
)

parser.add_argument(
    "SEAMM",
    "--initialize",
    group="dashboard options",
    action="store_true",
    help="initialize, or reinitialize, from the job files",
)
parser.add_argument(
    "SEAMM",
    "--no-check",
    group="dashboard options",
    action="store_true",
    help="do not check that jobs are in the database",
)

parser.add_argument(
    "SEAMM",
    "--console-log-level",
    group="dashboard options",
    default="INFO",
    choices=["CRITICAL", "ERROR", "WARNING", "INFO", "DEBUG", "NOTSET"],
    type=str.upper,
    help="the logging level for the dashboard console",
)

parser.add_argument(
    "SEAMM",
    "--log_dir",
    group="dashboard options",
    default="${root}/logs",
    action="store",
    help="The directory for logging",
)

# The rest of the options are for Flask, SQLAlchemy, Bootstrap, etc.
# and will be added to the Flask configuration.

# Flask options
parser.add_argument_group(
    "SEAMM",
    "flask options",
    "The options for Flask",
)

parser.add_argument(
    "SEAMM",
    "--env",
    group="flask options",
    default=argparse.SUPPRESS,
    help=(
        "What environment the app is running in. Flask and extensions may "
        "enable behaviors based on the environment, such as enabling debug "
        "mode. The env attribute maps to this config key. This is set by the "
        "FLASK_ENV environment variable and may not behave as expected if set "
        "in code."
        "\n"
        "Do not enable development when deploying in production."
        "\n"
        "Default: production"
    ),
)

parser.add_argument(
    "SEAMM",
    "--debug",
    group="flask options",
    default=argparse.SUPPRESS,
    help=(
        "Whether debug mode is enabled. When using flask run to start the "
        "development server, an interactive debugger will be shown for "
        "unhandled exceptions, and the server will be reloaded when code "
        "changes. The debug attribute maps to this config key. This is "
        "enabled when ENV is 'development' and is overridden by the "
        "FLASK_DEBUG environment variable. It may not behave as expected if "
        "set in code."
        "\n"
        "Do not enable debug mode when deploying in production."
        "\n"
        "Default: True if ENV is 'development', or False otherwise."
    ),
)

parser.add_argument(
    "SEAMM",
    "--testing",
    group="flask options",
    default=argparse.SUPPRESS,
    help=(
        "Enable testing mode. Exceptions are propagated rather than handled "
        "by the the app’s error handlers. Extensions may also change their "
        "behavior to facilitate easier testing. You should enable this in "
        "your own tests."
        "\n"
        "Default: False"
    ),
)

parser.add_argument(
    "SEAMM",
    "--propagate-exceptions",
    group="flask options",
    default=argparse.SUPPRESS,
    help=(
        "Exceptions are re-raised rather than being handled by the app’s "
        "error handlers. If not set, this is implicitly true if TESTING or "
        "DEBUG is enabled."
        "\n"
        "Default = None"
    ),
)

parser.add_argument(
    "SEAMM",
    "--preserve-context-on-exception",
    group="flask options",
    default=argparse.SUPPRESS,
    help=(
        "Don't pop the request context when an exception occurs. If not set, "
        "this is true if DEBUG is true. This allows debuggers to introspect "
        "the request data on errors, and should normally not need to be set "
        "directly."
        "\n"
        "Default = None"
    ),
)

parser.add_argument(
    "SEAMM",
    "--trap-http-exceptions",
    group="flask options",
    default=argparse.SUPPRESS,
    help=(
        "If there is no handler for an HTTPException-type exception, re-raise "
        "it to be handled by the interactive debugger instead of returning it "
        "as a simple error response."
        "\n"
        "Default = False"
    ),
)

parser.add_argument(
    "SEAMM",
    "--trap-bad-request-errors",
    group="flask options",
    default=argparse.SUPPRESS,
    help=(
        "Trying to access a key that doesn’t exist from request dicts like "
        "args and form will return a 400 Bad Request error page. Enable this "
        "to treat the error as an unhandled exception instead so that you get "
        "the interactive debugger. This is a more specific version of "
        "TRAP_HTTP_EXCEPTIONS. If unset, it is enabled in debug mode."
        "\n"
        "Default = None"
    ),
)

parser.add_argument(
    "SEAMM",
    "--secret-key",
    group="flask options",
    default=argparse.SUPPRESS,
    help=(
        "A secret key that will be used for securely signing the session "
        "cookie and can be used for any other security related needs by "
        "extensions or your application. It should be a long random string of "
        "bytes, although unicode is accepted too. For example, copy the "
        "output of this to your config:"
        "\n"
        "$ python -c 'import os; print(os.urandom(16))'"
        "b'_5#y2L\"F4Q8z\n\xec]/'"
        "\n"
        "Do not reveal the secret key when posting questions or committing "
        "code."
        "\n"
        "Default = None"
    ),
)

parser.add_argument(
    "SEAMM",
    "--jwt-secret-key",
    group="flask options",
    default=argparse.SUPPRESS,
    help=(
        """
        A secret key for the JWT token.
        """
    ),
)

parser.add_argument(
    "SEAMM",
    "--session-cookie-name",
    group="flask options",
    default=argparse.SUPPRESS,
    help=(
        "The name of the session cookie. Can be changed in case you already "
        "have a cookie with the same name."
        "\n"
        "Default = session"
    ),
)

parser.add_argument(
    "SEAMM",
    "--session-cookie-domain",
    group="flask options",
    default=argparse.SUPPRESS,
    help=(
        "The domain match rule that the session cookie will be valid for. If "
        "not set, the cookie will be valid for all subdomains of SERVER_NAME. "
        "If False, the cookie’s domain will not be set."
        "\n"
        "Default = None"
    ),
)

parser.add_argument(
    "SEAMM",
    "--session-cookie-path",
    group="flask options",
    default=argparse.SUPPRESS,
    help=(
        "The path that the session cookie will be valid for. If not set, the "
        "cookie will be valid underneath APPLICATION_ROOT or / if that is not "
        "set."
        "\n"
        "Default = None"
    ),
)

parser.add_argument(
    "SEAMM",
    "--session-cookie-secure",
    group="flask options",
    default=argparse.SUPPRESS,
    help=(
        "Browsers will only send cookies with requests over HTTPS if the "
        "cookie is marked “secure”. The application must be served over HTTPS "
        "for this to make sense."
        "\n"
        "Default = False"
    ),
)

parser.add_argument(
    "SEAMM",
    "--session-cookie-samesite",
    group="flask options",
    default="Lax",
    help=(
        "Restrict how cookies are sent with requests from external sites. Can "
        "be set to 'Lax' (recommended) or 'Strict'. See Set-Cookie options."
        "\n"
        "Default = 'Lax'"
    ),
)

parser.add_argument(
    "SEAMM",
    "--permanent-session-lifetime",
    group="flask options",
    default=argparse.SUPPRESS,
    help=(
        "If session.permanent is true, the cookie’s expiration will be set "
        "this number of seconds in the future. Can either be a "
        "datetime.timedelta or an int."
        "\n"
        "Flask’s default cookie implementation validates that the "
        "cryptographic signature is not older than this value."
        "\n"
        "Default: datetime.timedelta(days=31)"
    ),
)

parser.add_argument(
    "SEAMM",
    "--session-refresh-each-request",
    group="flask options",
    default=argparse.SUPPRESS,
    help=(
        "Control whether the cookie is sent with every response when "
        "session.permanent is true. Sending the cookie every time (the "
        "default) can more reliably keep the session from expiring, but uses "
        "more bandwidth. Non-permanent sessions are not affected."
        "\n"
        "Default = True"
    ),
)

parser.add_argument(
    "SEAMM",
    "--use-x-sendfile",
    group="flask options",
    default=argparse.SUPPRESS,
    help=(
        "When serving files, set the X-Sendfile header instead of serving the "
        "data with Flask. Some web servers, such as Apache, recognize this "
        "and serve the data more efficiently. This only makes sense when "
        "using such a server."
        "\n"
        "Default = False"
    ),
)

parser.add_argument(
    "SEAMM",
    "--send-file-max-age-default",
    group="flask options",
    default=argparse.SUPPRESS,
    help=(
        "When serving files, set the cache control max age to this number of "
        "seconds. Can either be a datetime.timedelta or an int. Override this "
        "value on a per-file basis using get_send_file_max_age() on the "
        "application or blueprint."
        "\n"
        "Default: timedelta(hours=12) (43200 seconds)"
    ),
)

parser.add_argument(
    "SEAMM",
    "--server-name",
    group="flask options",
    default=argparse.SUPPRESS,
    help=(
        "Inform the application what host and port it is bound to. Required "
        "for subdomain route matching support."
        "\n"
        "If set, will be used for the session cookie domain if "
        "SESSION_COOKIE_DOMAIN is not set. Modern web browsers will not allow "
        "setting cookies for domains without a dot. To use a domain locally, "
        "add any names that should route to the app to your hosts file."
        "\n"
        "127.0.0.1 localhost.dev"
        "If set, url_for can generate external URLs with only an application "
        "context instead of a request context."
        "\n"
        "Default: None"
    ),
)

parser.add_argument(
    "SEAMM",
    "--application-root",
    group="flask options",
    default=argparse.SUPPRESS,
    help=(
        "Inform the application what path it is mounted under by the "
        "application / web server. This is used for generating URLs outside "
        "the context of a request (inside a request, the dispatcher is "
        "responsible for setting SCRIPT_NAME instead; see Application "
        "Dispatching for examples of dispatch configuration)."
        "\n"
        "Will be used for the session cookie path if SESSION_COOKIE_PATH is "
        "not set."
        "\n"
        "Default = '/'"
    ),
)

parser.add_argument(
    "SEAMM",
    "--preferred-url-scheme",
    group="flask options",
    default=argparse.SUPPRESS,
    help=(
        "Use this scheme for generating external URLs when not in a request "
        "context."
        "\n"
        "default='http'"
    ),
)

parser.add_argument(
    "SEAMM",
    "--max-content-length",
    group="flask options",
    default=argparse.SUPPRESS,
    help=(
        "Don't read more than this many bytes from the incoming request data. "
        "If not set and the request does not specify a CONTENT_LENGTH, no "
        "data will be read for security."
        "\n"
        "Default: None"
    ),
)

parser.add_argument(
    "SEAMM",
    "--json-as-ascii",
    group="flask options",
    default=argparse.SUPPRESS,
    help=(
        "Serialize objects to ASCII-encoded JSON. If this is disabled, the "
        "JSON will be returned as a Unicode string, or encoded as UTF-8 by "
        "jsonify. This has security implications when rendering the JSON into "
        "JavaScript in templates, and should typically remain enabled."
        "\n"
        "default=True"
    ),
)

parser.add_argument(
    "SEAMM",
    "--json-sort-keys",
    group="flask options",
    default=argparse.SUPPRESS,
    help=(
        "Sort the keys of JSON objects alphabetically. This is useful for "
        "caching because it ensures the data is serialized the same way no "
        "matter what Python’s hash seed is. While not recommended, you can "
        "disable this for a possible performance improvement at the cost of "
        "caching."
        "\n"
        "default=True"
    ),
)

parser.add_argument(
    "SEAMM",
    "--jsonify-prettyprint-regular",
    group="flask options",
    default=argparse.SUPPRESS,
    help=(
        "jsonify responses will be output with newlines, spaces, and "
        "indentation for easier reading by humans. Always enabled in debug "
        "mode."
        "\n"
        "default=False"
    ),
)

parser.add_argument(
    "SEAMM",
    "--jsonify-mimetype",
    group="flask options",
    default=argparse.SUPPRESS,
    help=("The mimetype of jsonify responses." "\n" "default='application/json'"),
)

parser.add_argument(
    "SEAMM",
    "--templates-auto-reload",
    group="flask options",
    default=argparse.SUPPRESS,
    help=(
        "Reload templates when they are changed. If not set, it will be "
        "enabled in debug mode."
        "\n"
        "Default: None"
    ),
)

parser.add_argument(
    "SEAMM",
    "--explain-template-loading",
    group="flask options",
    default=argparse.SUPPRESS,
    help=(
        "Log debugging information tracing how a template file was loaded. "
        "This can be useful to figure out why a template was not loaded or "
        "the wrong file appears to be loaded."
        "\n"
        "default=False"
    ),
)

parser.add_argument(
    "SEAMM",
    "--max-cookie-size",
    group="flask options",
    default=argparse.SUPPRESS,
    help=(
        "Warn if cookie headers are larger than this many bytes. Defaults to "
        "4093. Larger cookies may be silently ignored by browsers. Set to 0 "
        "to disable the warning."
        "\n"
        "default=4093"
    ),
)

# SQLAlchemy options
parser.add_argument_group(
    "SEAMM",
    "sqlalchemy options",
    "The SQLAlchemy options for the dashboard",
)

parser.add_argument(
    "SEAMM",
    "--sqlalchemy-database-uri",
    group="sqlalchemy options",
    default="",
    help=(
        "The database URI that should be used for the connection. Examples:"
        "\n"
        "sqlite:////tmp/test.db"
        "\n"
        "mysql://username:password@server/db"
    ),
)

parser.add_argument(
    "SEAMM",
    "--sqlalchemy-binds",
    group="sqlalchemy options",
    default=argparse.SUPPRESS,
    help=(
        "A dictionary that maps bind keys to SQLAlchemy connection URIs. For "
        "more information about binds see Multiple Databases with Binds."
    ),
)

parser.add_argument(
    "SEAMM",
    "--sqlalchemy-echo",
    group="sqlalchemy options",
    default=argparse.SUPPRESS,
    help=(
        "If set to True SQLAlchemy will log all the statements issued to "
        "stderr which can be useful for debugging."
    ),
)

parser.add_argument(
    "SEAMM",
    "--sqlalchemy-record-queries",
    group="sqlalchemy options",
    default=argparse.SUPPRESS,
    help=(
        "Can be used to explicitly disable or enable query recording. Query "
        "recording automatically happens in debug or testing mode. See "
        "get_debug_queries() for more information."
    ),
)

parser.add_argument(
    "SEAMM",
    "--sqlalchemy-track-modifications",
    group="sqlalchemy options",
    default=False,
    help=(
        "If set to True, Flask-SQLAlchemy will track modifications of objects "
        "and emit signals. The default is None, which enables tracking but "
        "issues a warning that it will be disabled by default in the future. "
        "This requires extra memory and should be disabled if not needed."
        "\n"
        "default=False"
    ),
)

parser.add_argument(
    "SEAMM",
    "--sqlalchemy-engine-options",
    group="sqlalchemy options",
    default=argparse.SUPPRESS,
    help=(
        "A dictionary of keyword args to send to create_engine(). See also "
        "engine_options to SQLAlchemy."
    ),
)

# Bootstrap options
parser.add_argument_group(
    "SEAMM",
    "bootstrap options",
    "The Bootstrap options for the dashboard",
)

parser.add_argument(
    "SEAMM",
    "--bootstrap-use-minified",
    group="bootstrap options",
    default=argparse.SUPPRESS,
    help=(
        "Whether or not to use the minified versions of the css/js files."
        "\n"
        "default=True"
    ),
)

parser.add_argument(
    "SEAMM",
    "--bootstrap-serve-local",
    group="bootstrap options",
    default=argparse.SUPPRESS,
    help=(
        "If True, Bootstrap resources will be served from the local app "
        "instance every time. See CDN support for details."
        "\n"
        "default=False"
    ),
)

parser.add_argument(
    "SEAMM",
    "--bootstrap-local-subdomain",
    group="bootstrap options",
    default=argparse.SUPPRESS,
    help=(
        "Passes a subdomain parameter to the generated Blueprint. Useful "
        "when serving assets locally from a different subdomain."
    ),
)

parser.add_argument(
    "SEAMM",
    "--bootstrap-cdn-force-ssl",
    group="bootstrap options",
    default=argparse.SUPPRESS,
    help=(
        "If a CDN resource url starts with //, prepend 'https:' to it."
        "\n"
        "default=True"
    ),
)

parser.add_argument(
    "SEAMM",
    "--bootstrap-querystring-revving",
    group="bootstrap options",
    default=argparse.SUPPRESS,
    help=(
        "If True, will append a querystring with the current version to all "
        "static resources served locally. This ensures that upon upgrading "
        "Flask-Bootstrap, these resources are refreshed."
        "\n"
        "default=True"
    ),
)

# Mail arguments - for flask mail
# https://pythonhosted.org/Flask-Mail/
parser.add_argument_group(
    "SEAMM",
    "mail options",
    "The FlaskMail options for the dashboard",
)

parser.add_argument(
    "SEAMM",
    "--mail-server",
    group="mail options",
    default=argparse.SUPPRESS,
    help="default=localhost",
)

parser.add_argument(
    "SEAMM",
    "--mail-port",
    group="mail options",
    default=argparse.SUPPRESS,
    help="default=25",
)

parser.add_argument(
    "SEAMM",
    "--mail-use-TLS",
    group="mail options",
    default=argparse.SUPPRESS,
    help="default=False",
)

parser.add_argument(
    "SEAMM",
    "--mail-use-SSL",
    group="mail options",
    default=argparse.SUPPRESS,
    help="default=False",
)

parser.add_argument(
    "SEAMM",
    "--mail-debug",
    group="mail options",
    default=argparse.SUPPRESS,
    help="default=app.debug",
)

parser.add_argument(
    "SEAMM",
    "--mail-username",
    group="mail options",
    default=argparse.SUPPRESS,
    help="default=None",
)

parser.add_argument(
    "SEAMM",
    "--mail-password",
    group="mail options",
    default=argparse.SUPPRESS,
    help="default=None",
)

parser.add_argument(
    "SEAMM",
    "--mail-default-sender",
    group="mail options",
    default=argparse.SUPPRESS,
    help="default=None",
)

parser.add_argument(
    "SEAMM",
    "--mail-max-emails",
    group="mail options",
    default=argparse.SUPPRESS,
    help="default=None",
)

parser.add_argument(
    "SEAMM",
    "--mail-suppress-send",
    group="mail options",
    default=argparse.SUPPRESS,
    help="default=app.testing",
)

parser.add_argument(
    "SEAMM",
    "--mail-ascii-attachments",
    group="mail options",
    default=argparse.SUPPRESS,
    help="default=False",
)

parser.add_argument(
    "SEAMM",
    "--mail-subject-prefix",
    group="mail options",
    default=argparse.SUPPRESS,
    help="",
)

parser.add_argument(
    "SEAMM", "--mail-sender", group="mail options", default=argparse.SUPPRESS, help=""
)

# And handle the command-line arguments and ini file options.
# Working around pytest :-(
if "pytest" in sys.argv[0]:
    parser.parse_args([])
else:
    parser.parse_args()
options = parser.get_options("SEAMM")

# Fix the database uri
if "sqlalchemy_database_uri" in options and options["sqlalchemy_database_uri"] == "":
    path = Path(options["datastore"]).expanduser().resolve() / "seamm.db"
    options["sqlalchemy_database_uri"] = f"sqlite:///{str(path)}"

if __name__ == "__main__":
    import pprint

    # options, unknown = parser.parse_known_args()

    pprint.pprint(vars(options))



import logging
import logging.handlers
from pathlib import Path


def setup_logging(datastore, options):
    """
    Sets up logging to console (INFO+) and logging of log file
    logs/myapp-<timestamp>.log. You can create a an extra logger to represent
    areas in your app: logger1 = logging.getLogger('area1')

    Logging Levels:

    https://docs.python.org/3/howto/logging.html#logging-levels

    DEBUG (default for FILE): Detailed information, typically of interest only
        when diagnosing problems.
    INFO (default for CONSOLE): Confirmation that things are working as
        expected.
    WARNING: An indication that something unexpected happened, or indicative of
        some problem in the near future (e.g. 'disk space low').
        The software is still working as expected.
    ERROR: Due to a more serious problem, the software has not been able
        to perform some function.
    CRITICAL: A serious error, indicating that the program itself may be unable
        to continue running.

    Params
        options
    """

    # Make sure the logs folder exists (avoid FileNotFoundError)
    log_dir = Path(options["log_dir"].replace("%datastore%", datastore)).expanduser()

    if not log_dir.is_dir():
        log_dir.mkdir(parents=True, exist_ok=True)

    # Set up logging to a file (overwriting)
    log_file = log_dir / "dashboard.log"

    # Get root logger
    logger = logging.getLogger()
    logger.setLevel("DEBUG")

    # Create a handler that writes INFO messages or higher to sys.stderr
    console_handler = logging.StreamHandler()
    console_handler.setLevel(options["console_log_level"])
    console_formatter = logging.Formatter("%(name)s:%(levelname)s:%(message)s")
    console_handler.setFormatter(console_formatter)
    logger.addHandler(console_handler)

    # And one to log in files
    file_handler = logging.handlers.TimedRotatingFileHandler(
        str(log_file), when="W6", backupCount=4
    )
    file_handler.setLevel(options["log_level"])
    file_formatter = logging.Formatter("%(asctime)s %(name)s:%(levelname)s:%(message)s")
    file_handler.setFormatter(file_formatter)
    logger.addHandler(file_handler)

    new_logger = logging.getLogger("dashboard")
    new_logger.info(
        "Logging to the console at level {}.".format(options["console_log_level"])
    )
    new_logger.info("Logging to {} at level {}.".format(log_file, options["log_level"]))

    # Demo usage
    # logger = logging.getLogger(__name__)
    # logger.debug('a debug log message')
    # logger.info('an info log message')
    # logger.warning('a warning log message')
    # logger.error('an error log message')
    # logger.critical('a critical log message')



"""
A patch so that we only need csrf tokens from flask-jwt-extended with specified
routes.

This allows two things:
    1. Using jwts on all routes to identify users
    2. Using csrf tokens from the JWT on API routes, while front-end views use CSRF
       tokens which are part of flask-wtforms
"""


from flask import request

from flask_jwt_extended.config import config
from flask_jwt_extended.exceptions import (
    CSRFError,
    NoAuthorizationError,
)

from flask import current_app

import flask_jwt_extended


def _decode_jwt_from_cookies(refresh):
    if refresh:
        cookie_key = config.refresh_cookie_name
        csrf_header_key = config.refresh_csrf_header_name
        csrf_field_key = config.refresh_csrf_field_name
    else:
        cookie_key = config.access_cookie_name
        csrf_header_key = config.access_csrf_header_name
        csrf_field_key = config.access_csrf_field_name

    encoded_token = request.cookies.get(cookie_key)
    if not encoded_token:
        raise NoAuthorizationError('Missing cookie "{}"'.format(cookie_key))

    if (
        config.csrf_protect
        and request.method in config.csrf_request_methods
        and current_app.config["JWT_CSRF_ACCESS_PATH"] in request.url
    ):
        csrf_value = request.headers.get(csrf_header_key, None)
        if not csrf_value and config.csrf_check_form:
            csrf_value = request.form.get(csrf_field_key, None)
        if not csrf_value:
            raise CSRFError("Missing CSRF token")
    else:
        csrf_value = None

    return encoded_token, csrf_value


flask_jwt_extended.view_decorators._decode_jwt_from_cookies = _decode_jwt_from_cookies



from copy import deepcopy

from flask import render_template, url_for, request, flash, redirect
from flask_jwt_extended import jwt_required, get_current_user

from wtforms import BooleanField

from . import projects
from .forms import EditProject, ManageProjectAccessForm

from seamm_datastore.database.models import Project, User, UserProjectAssociation

from seamm_dashboard import authorize, db


def _bind_users_to_form(form, current_user, project_id):

    actions = ["read", "update", "create", "delete", "manage"]
    users = User.query.all()
    user_names = []

    if len(users) > 1:
        # Put current user first
        users.remove(current_user)
        reorder_users = [current_user]
        reorder_users.extend(users)
        users = reorder_users

    for user in users:
        field_name = f"user_{user.id}"
        permissions = []
        assoc = UserProjectAssociation.query.filter_by(
            resource_id=project_id, entity_id=user.id
        ).one_or_none()

        if assoc:
            permissions = assoc.permissions

        for action in actions:
            checked = action in permissions
            setattr(form, f"{field_name}_{action}", BooleanField(default=checked))

        user_names.append({"username": user.username, "id": user.id})

    return form, user_names


@projects.route("/views/projects")
def project_list():
    return render_template("projects/project_list.html")


@projects.route("/views/projects/<id>/jobs")
@projects.route("/views//projects/<id>/jobs")
@jwt_required(optional=True)
def project_jobs_list(id):

    project = Project.query.get(id)

    manage_project = authorize.manage(project)
    edit_project = authorize.update(project)

    # Build the url ourselves.
    base_url = url_for("main.index")
    edit_url = base_url + f"projects/{id}/edit"
    manage_url = base_url + f"projects/{id}/manage"

    return render_template(
        "jobs/jobs_list.html",
        project=True,
        manage_project=manage_project,
        edit_project=edit_project,
        edit_url=edit_url,
        manage_url=manage_url,
    )


@projects.route("/projects/<project_id>/edit", methods=["GET", "POST"])
@jwt_required(optional=True)
def edit_project(project_id):

    project = Project.query.get(project_id)

    if not authorize.update(project):
        return render_template("401.html")

    form = EditProject()

    # Build the url ourselves.
    base_url = url_for("main.index")
    project_url = base_url + f"#projects/{project_id}/jobs"

    if form.validate_on_submit():
        project.name = form.name.data
        project.description = form.notes.data
        db.session.commit()
        flash("Project updated successfully.", "successs")

        return redirect(project_url)

    elif request.method == "GET":
        form.name.data = project.name
        form.notes.data = project.description

    return render_template(
        "jobs/edit_job.html",
        title=f"Edit Project {project_id}",
        form=form,
        back_url=project_url,
    )


@projects.route("/projects/<project_id>/manage", methods=["GET", "POST"])
@jwt_required(optional=True)
def manage_project(project_id):

    project = Project.query.get(project_id)

    if not project:
        return render_template("404.html")

    if not authorize.manage(project):
        return render_template("401.html")

    form = deepcopy(ManageProjectAccessForm)

    form, usernames = _bind_users_to_form(
        form, current_user=get_current_user(), project_id=project.id
    )

    form = form()

    # Build the url ourselves.
    base_url = url_for("main.index")
    project_url = base_url + f"#projects/{project_id}/jobs"

    if request.method == "POST":
        if form.validate_on_submit():

            user_keys = [
                x for x in form.data.keys() if "user" in x if form.data[x] is True
            ]

            permissions_dict = {}

            for key in user_keys:
                split = key.split("_")
                user_id = int(split[1])
                permission = split[2]

                try:
                    permissions_dict[user_id].append(permission)
                except KeyError:
                    permissions_dict[user_id] = [permission]

            # Find an entries for special user permissions which exist for this project
            users = User.query.all()

            for entry in users:
                user_id = entry.id

                try:
                    permissions = permissions_dict[user_id]
                except KeyError:
                    permissions = []

                assoc = UserProjectAssociation.query.filter_by(
                    entity_id=user_id, resource_id=project_id
                ).one_or_none()

                if not assoc:
                    assoc = UserProjectAssociation(
                        entity_id=user_id,
                        resource_id=project_id,
                        permissions=permissions,
                    )
                else:
                    assoc.permissions = permissions

                db.session.add(assoc)
                db.session.commit()

            flash(f"Permissions for {project.name} successfully updated.")
            return redirect(project_url)

    return render_template(
        "projects/project_access.html",
        form=form,
        users=usernames,
        project=project,
        back_url=project_url,
    )
